﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using Microsoft.Xna.Framework;

namespace NetworkCommon
{
  /// <summary>
  /// Contains data about a single color entry in a color sequence
  /// </summary>
  public class ColorSeqEntry
  {
    /// <summary>
    /// Creates a color sequence entry from the specified byte array
    /// that contains the data portion of a received network message
    /// </summary>
    /// <param name="data">
    /// data portion of the message
    /// </param>
    /// <param name="offset">
    /// starting index of data to start extracting a entry.  This will be
    /// updated to account for the amount read by this function
    /// </param>
    static public ColorSeqEntry Create(byte[] data,ref int offset)
    {
      // load the color
      Vector3 color;
      color.X = BitConverter.ToSingle(data,offset);
      offset += sizeof(float);
      color.Y = BitConverter.ToSingle(data,offset);
      offset += sizeof(float);
      color.Z = BitConverter.ToSingle(data,offset);
      offset += sizeof(float);
      
      // load the life
      float life = BitConverter.ToSingle(data,offset);
      offset += sizeof(float);
      
      // create the entry
      return new ColorSeqEntry(color,life);
    }
    
    /// <summary>
    /// Creates a new color sequence entry
    /// </summary>
    /// <param name="color">
    /// Color to use for the entry
    /// </param>
    /// <param name="life">
    /// Percent of a particle's life, in the range (0,1], that the color
    /// should be used to color a particle
    /// </param>
    public ColorSeqEntry(Vector3 color,float life)
    {
      Color = color;
      Life  = life;
    }
    
    /// <summary>
    /// Writes the entry's data to the binary writer
    /// </summary>
    /// <param name="writer">
    /// where to write the data to
    /// </param>
    /// <returns>
    /// number of bytes written to writer
    /// </returns>
    public int WriteContents(BinaryWriter writer)
    {
      int back = 0;
      
      // write the color
      writer.Write(Color.X);
      back += sizeof(float);
      writer.Write(Color.Y);
      back += sizeof(float);
      writer.Write(Color.Z);
      back += sizeof(float);
      
      // write the percentage
      writer.Write(Life);
      back += sizeof(float);
      
      return back;
    }
    
    /// <summary>
    /// Color of the entry
    /// </summary>
    public Vector3 Color { get; set; }
    
    /// <summary>
    /// Percent of a particle's life, in the range [0,1], that the color
    /// should be used to color a particle
    /// </summary>
    public float Life { get; set; }
  }
}
