﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using Microsoft.Xna.Framework;

namespace NetworkCommon
{
  /// <summary>
  /// Contains data about a single color entry in a color set
  /// </summary>
  public class ColorSetEntry
  {
    /// <summary>
    /// Creates a color set entry from the specified byte array
    /// that contains the data portion of a received network message
    /// </summary>
    /// <param name="data">
    /// data portion of the message
    /// </param>
    /// <param name="offset">
    /// starting index of data to start extracting a entry.  This will be
    /// updated to account for the amount read by this function
    /// </param>
    static public ColorSetEntry Create(byte[] data,ref int offset)
    {
      // load the color
      Vector3 color;
      color.X = BitConverter.ToSingle(data,offset);
      offset += sizeof(float);
      color.Y = BitConverter.ToSingle(data,offset);
      offset += sizeof(float);
      color.Z = BitConverter.ToSingle(data,offset);
      offset += sizeof(float);
      
      // load the chance
      float chance = BitConverter.ToSingle(data,offset);
      offset += sizeof(float);
      
      // create the entry
      return new ColorSetEntry(color,chance);
    }
    
    /// <summary>
    /// Creates a new color set entry
    /// </summary>
    /// <param name="color">
    /// Color to use for the entry
    /// </param>
    /// <param name="prob">
    /// Percent chance, in the range [0,1], of the color being selected for
    /// the color set
    /// </param>
    public ColorSetEntry(Vector3 color,float prob)
    {
      Color  = color;
      Chance = prob;
    }
    
    /// <summary>
    /// Writes the entry's data to the binary writer
    /// </summary>
    /// <param name="writer">
    /// where to write the data to
    /// </param>
    /// <returns>
    /// number of bytes written to writer
    /// </returns>
    public int WriteContents(BinaryWriter writer)
    {
      int back = 0;
      
      // write the color
      writer.Write(Color.X);
      back += sizeof(float);
      writer.Write(Color.Y);
      back += sizeof(float);
      writer.Write(Color.Z);
      back += sizeof(float);
      
      // write the percentage
      writer.Write(Chance);
      back += sizeof(float);
      
      return back;
    }
    
    /// <summary>
    /// Color of the entry
    /// </summary>
    public Vector3 Color { get; set; }
    
    /// <summary>
    /// Percent chance, in the range [0,1], of the color being selected for
    /// the color set
    /// </summary>
    public float Chance { get; set; }
  }
}
