﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;

namespace NetworkCommon
{
  /// <summary>
  /// Base class for network messages
  /// </summary>
  abstract public class Message
  {
    /// <summary>
    /// Static constructor for Message that takes care of populating the
    /// dictionary of message types to delegates that create instances of that
    /// type
    /// </summary>
    static Message()
    {
      // create the dictionary
      type_to_create = new Dictionary<MessageIDs,CreateMessage>();
      
      // add the various type/delegate mappings to the dictionary
      type_to_create.Add(MessageIDs.CONNECTION_REJECTED,
        (byte[] data) => new MsgConnectionRejected(data));
      type_to_create.Add(MessageIDs.CLOSE_CONNECTION,
        (byte[] data) => new MsgCloseConnection(data));
      type_to_create.Add(MessageIDs.SET_SINGLE,
        (byte[] data) => new MsgSetSingle(data));
      type_to_create.Add(MessageIDs.SET_STRING,
        (byte[] data) => new MsgSetString(data));
      type_to_create.Add(MessageIDs.COLOR_SET,
        (byte[] data) => new MsgColorSet(data));
      type_to_create.Add(MessageIDs.COLOR_SEQ,
        (byte[] data) => new MsgColorSeq(data));
    }
    
    /// <summary>
    /// Receives a message from the stream
    /// </summary>
    /// <param name="stream">
    /// stream to read a message from
    /// </param>
    /// <returns>
    /// message that was received by the stream
    /// </returns>
    static public Message Receive(Stream stream)
    {
      // create a binary reader for the stream
      BinaryReader reader = new BinaryReader(stream);
      
      // check the version number
      int version = reader.ReadInt32();
      if (version != 1)
      {
        throw new InvalidDataException(
          "Wrong version.  Wanted \"1\", found \"" + version + "\"");
      }
      
      // get the length
      int length = reader.ReadInt32();
      if (length < 0)
      {
        throw new InvalidDataException("Invalid length of \"" + length + "\"");
      }
      
      // get the message id
      int id = reader.ReadInt32();
      if (id <= (int)MessageIDs.MSG_ID_MIN)
      {
        throw new InvalidDataException(
          "Message id below the min valid id.  Min is "
          + ((int)MessageIDs.MSG_ID_MIN + 1) + ", found " + id);
      }
      else if (id >= (int)MessageIDs.MSG_ID_MAX)
      {
        throw new InvalidDataException(
          "Message id above the max valid id.  Max is "
          + ((int)MessageIDs.MSG_ID_MAX - 1) + ", found " + id);
      }
      
      // get the message data
      byte[] data = new byte[length];
      int bytes_read = 0;
      while (bytes_read < length)
      {
        byte[] portion = reader.ReadBytes(length - bytes_read);
        Array.Copy(portion,0,data,bytes_read,portion.Length);
        
        bytes_read += portion.Length;
      }
      
      // create the appropriate message instance for the type
      if (type_to_create.ContainsKey((MessageIDs)id))
      {
        CreateMessage create = type_to_create[(MessageIDs)id];
        return create(data);
      }
      else
      {
        throw new NotImplementedException("No creation entry for id " + id);
      }
    }

    /// <summary>
    /// Writes the mssage to the stream in binary form
    /// </summary>
    /// <param name="stream">
    /// Stream to write the message to
    /// </param>
    public void Send(Stream stream)
    {
      using (MemoryStream mem = new MemoryStream())
      using (BinaryWriter writer = new BinaryWriter(mem))
      {
        int length = 0;
        
        // version
        writer.Write(1);
        
        // length (placeholder value that will be set after the subtype fills
        // in the actual message contents)
        writer.Write(0);
        
        // message id
        writer.Write((int)GetMessageID());
        
        // add in the subtype's data
        length = WriteContents(writer);
        
        // determine the value the length field should have
        writer.Seek(sizeof(int),SeekOrigin.Begin);
        writer.Write(length);
        
        // now that the message is created, send it to the stream
        byte[] data = mem.ToArray();
        stream.Write(data,0,data.Length);
      }
    }
    
    /// <summary>
    /// Retrieves the subtype's message id
    /// </summary>
    /// <returns>
    /// message id of the subtype
    /// </returns>
    abstract public MessageIDs GetMessageID();
    
    /// <summary>
    /// Puts the subtype's data into the message via the specified binary
    /// writer
    /// </summary>
    /// <param name="writer">
    /// where to write the subtype's data to
    /// </param>
    /// <returns>
    /// number of bytes written to writer
    /// </returns>
    abstract protected int WriteContents(BinaryWriter writer);

    /// <summary>
    /// Delegate type for creating instances of Message's subtypes.
    /// </summary>
    /// <param name="data">message contents portion of a login message</param>
    /// <returns></returns>
    private delegate Message CreateMessage(byte[] data);

    /// <summary>
    /// Dictionary of message types to delegates that create instances of that
    /// type
    /// </summary>
    static private Dictionary<MessageIDs,CreateMessage> type_to_create;
  }
}
