﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;

namespace NetworkCommon
{
  /// <summary>
  /// Passes an entire color set
  /// </summary>
  public class MsgColorSet : Message
  {
    /// <summary>
    /// Creates a color set message
    /// </summary>
    /// <param name="val">
    /// color set entries that make up the color set
    /// </param>
    public MsgColorSet(ColorSetEntry[] val)
    {
      Value = val;
    }
    
    /// <summary>
    /// Creates a color set message from the specified byte array
    /// that contains the data portion of a received network message
    /// </summary>
    /// <param name="data">
    /// data portion of the message
    /// </param>
    public MsgColorSet(byte[] data)
    {
      if (data.Length < sizeof(int))
      {
        throw new ArgumentException("Invalid length of data");
      }
      
      // position in data for what has been extracted so far
      int pos = 0;
      
      // extract the number of entries
      int num_entries = BitConverter.ToInt32(data,pos);
      pos += sizeof(int);
      if (num_entries < 0)
      {
        throw new ArgumentException("Number of entries is negative.  Got "
          + num_entries);
      }
      
      // create the storage for the entries
      Value = new ColorSetEntry[num_entries];
      
      // extract the entries
      for (int i = 0; i < num_entries; ++i)
      {
        Value[i] = ColorSetEntry.Create(data,ref pos);
      }
    }
    
    /// <summary>
    /// Retrieves the message id
    /// </summary>
    /// <returns>
    /// message id
    /// </returns>
    override public MessageIDs GetMessageID()
    {
      return MessageIDs.COLOR_SET;
    }
    
    /// <summary>
    /// Puts the any needed data into the message via the specified binary
    /// writer
    /// </summary>
    /// <param name="writer">
    /// where to write the data to
    /// </param>
    /// <returns>
    /// number of bytes written to writer
    /// </returns>
    override protected int WriteContents(BinaryWriter writer)
    {
      int back = 0;
      
      // write the number of entries
      writer.Write(Value.Length);
      back += sizeof(int);
      
      // write the entries
      foreach (ColorSetEntry entry in Value)
      {
        back += entry.WriteContents(writer);
      }
      
      return back;
    }
    
    /// <summary>
    /// String value contained in the message
    /// </summary>
    public ColorSetEntry[] Value { get; private set; }
  }
}
