﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;

namespace NetworkCommon
{
  /// <summary>
  /// Passes a single numeric value and the identifier for which value it is
  /// </summary>
  public class MsgSetSingle : Message
  {
    /// <summary>
    /// Creates a set single message
    /// </summary>
    /// <param name="id">
    /// id of the parameter the value is for
    /// </param>
    /// <param name="val">
    /// value to set the parameter to
    /// </param>
    public MsgSetSingle(ParameterIDs id,float val)
    {
      ID = id;
      Value = val;
    }
    
    /// <summary>
    /// Creates a set single message from the specified byte array
    /// that contains the data portion of a received network message
    /// </summary>
    /// <param name="data">
    /// data portion of the message
    /// </param>
    public MsgSetSingle(byte[] data)
    {
      if (data.Length != (sizeof(ParameterIDs) + sizeof(float)))
      {
        throw new ArgumentException("Invalid length of data");
      }
      
      // extract the parameter id
      int param_id = BitConverter.ToInt32(data,0);
      if (param_id <= (int)ParameterIDs.PARAM_ID_MIN)
      {
        throw new ArgumentException("Param id below min.  Min is "
          + (ParameterIDs.PARAM_ID_MIN + 1) + ", got " + param_id);
      }
      else if (param_id >= (int)ParameterIDs.PARAM_ID_MAX)
      {
        throw new ArgumentException("Param id above max.  Max is "
          + (ParameterIDs.PARAM_ID_MAX - 1) + ", got " + param_id);
      }
      ID = (ParameterIDs)param_id;
      
      // extract the value
      Value = BitConverter.ToSingle(data,sizeof(ParameterIDs));
    }
    
    /// <summary>
    /// Retrieves the message id
    /// </summary>
    /// <returns>
    /// message id
    /// </returns>
    override public MessageIDs GetMessageID()
    {
      return MessageIDs.SET_SINGLE;
    }
    
    /// <summary>
    /// Puts the any needed data into the message via the specified binary
    /// writer
    /// </summary>
    /// <param name="writer">
    /// where to write the data to
    /// </param>
    /// <returns>
    /// number of bytes written to writer
    /// </returns>
    override protected int WriteContents(BinaryWriter writer)
    {
      // write the parameter id
      writer.Write((int)ID);
      
      // write the value
      writer.Write(Value);
      
      return sizeof(int) + sizeof(float);
    }
    
    /// <summary>
    /// Parameter id of the value being set
    /// </summary>
    public ParameterIDs ID { get; private set; }
    
    /// <summary>
    /// Numeric value contained in the message
    /// </summary>
    public float Value { get; private set; }
  }
}
