﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;

namespace NetworkCommon
{
  /// <summary>
  /// Passes a single string value and the identifier for which value it is
  /// </summary>
  public class MsgSetString : Message
  {
    /// <summary>
    /// Creates a set string message
    /// </summary>
    /// <param name="id">
    /// id of the parameter the value is for
    /// </param>
    /// <param name="val">
    /// value to set the parameter to
    /// </param>
    public MsgSetString(ParameterIDs id,string val)
    {
      ID = id;
      Value = val;
    }
    
    /// <summary>
    /// Creates a set string message from the specified byte array
    /// that contains the data portion of a received network message
    /// </summary>
    /// <param name="data">
    /// data portion of the message
    /// </param>
    public MsgSetString(byte[] data)
    {
      if (data.Length < (sizeof(ParameterIDs) + sizeof(int)))
      {
        throw new ArgumentException("Invalid length of data");
      }
      
      // position in data for what has been extracted so far
      int pos = 0;
      
      // extract the parameter id
      int param_id = BitConverter.ToInt32(data,pos);
      pos += sizeof(int);
      if (param_id <= (int)ParameterIDs.PARAM_ID_MIN)
      {
        throw new ArgumentException("Param id below min.  Min is "
          + (ParameterIDs.PARAM_ID_MIN + 1) + ", got " + param_id);
      }
      else if (param_id >= (int)ParameterIDs.PARAM_ID_MAX)
      {
        throw new ArgumentException("Param id above max.  Max is "
          + (ParameterIDs.PARAM_ID_MAX - 1) + ", got " + param_id);
      }
      ID = (ParameterIDs)param_id;
      
      // get the length of the string
      int len = BitConverter.ToInt32(data,pos);
      pos += sizeof(int);
      if (len < 0)
      {
        throw new ArgumentException("String length is negative.  Got " + len);
      }
      
      // get the string
      UTF8Encoding encoding = new UTF8Encoding();
      Value = encoding.GetString(data,pos,len);
    }
    
    /// <summary>
    /// Retrieves the message id
    /// </summary>
    /// <returns>
    /// message id
    /// </returns>
    override public MessageIDs GetMessageID()
    {
      return MessageIDs.SET_STRING;
    }
    
    /// <summary>
    /// Puts the any needed data into the message via the specified binary
    /// writer
    /// </summary>
    /// <param name="writer">
    /// where to write the data to
    /// </param>
    /// <returns>
    /// number of bytes written to writer
    /// </returns>
    override protected int WriteContents(BinaryWriter writer)
    {
      int back = 0;
      
      // write the parameter id
      writer.Write((int)ID);
      back += sizeof(int);
      
      // write the string length
      UTF8Encoding encoding = new UTF8Encoding();
      int len = encoding.GetByteCount(Value);
      writer.Write(len);
      back += sizeof(int);
      
      // write the value
      writer.Write(encoding.GetBytes(Value));
      back += len;
      
      return back;
    }
    
    /// <summary>
    /// Parameter id of the value being set
    /// </summary>
    public ParameterIDs ID { get; private set; }
    
    /// <summary>
    /// String value contained in the message
    /// </summary>
    public string Value { get; private set; }
  }
}
