﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace NetworkCommon
{
  /// <summary>
  /// Enum listing the various configuration parameters
  /// </summary>
  public enum ParameterIDs : int
  {
    /// <summary>
    /// not a parameter id, but a place holder for 1 below the lowest valid
    /// parameter id
    /// </summary>
    PARAM_ID_MIN,
    
    #region Camera field IDs
    /// <summary>
    /// X world coordinate of where the camera should be looking at
    /// </summary>
    CAM_LOOKAT_X,
    
    /// <summary>
    /// Y world coordinate of where the camera should be looking at
    /// </summary>
    CAM_LOOKAT_Y,
    
    /// <summary>
    /// Z world coordinate of where the camera should be looking at
    /// </summary>
    CAM_LOOKAT_Z,
    
    /// <summary>
    /// distance the camera's position should be from the look at position,
    /// must be a non-negative value
    /// </summary>
    CAM_ORBIT_RANGE,
    
    /// <summary>
    /// yaw value for determining the camera's position relative to the look
    /// at, in degrees in the range [-180,180]
    /// </summary>
    CAM_ORBIT_YAW,
    
    /// <summary>
    /// pitch value for determining the camera's positiion relative to the look
    /// at, in degrees in the range [-180,180]
    /// </summary>
    CAM_ORBIT_PITCH,
    #endregion

    #region Emitter field IDs
    /// <summary>
    /// Shape of the emitter
    /// </summary>
    EMITTER_SHAPE,
    
    /// <summary>
    /// X world coordinate position of the emitter
    /// </summary>
    EMITTER_POS_X,
    
    /// <summary>
    /// Y world coordinate position of the emitter
    /// </summary>
    EMITTER_POS_Y,
    
    /// <summary>
    /// Z world coordinate position of the emitter
    /// </summary>
    EMITTER_POS_Z,
    
    /// <summary>
    /// yaw, in degrees in the range [-180,180], to use for the emitter's
    /// rotation
    /// </summary>
    EMITTER_ROT_YAW,
    
    /// <summary>
    /// pitch, in degrees in the range [-180,180], to use for the emitter's
    /// rotation
    /// </summary>
    EMITTER_ROT_PITCH,
    
    /// <summary>
    /// roll, in degrees in the range [-180,180], to use for the emitter's
    /// rotation
    /// </summary>
    EMITTER_ROT_ROLL,
    
    /// <summary>
    /// amount in X that the emitter should be scaled by
    /// </summary>
    EMITTER_SCALE_X,
    
    /// <summary>
    /// amount in Y that the emitter should be scaled by
    /// </summary>
    EMITTER_SCALE_Y,
    
    /// <summary>
    /// amount in Z that the emitter should be scaled by
    /// </summary>
    EMITTER_SCALE_Z,
    
    /// <summary>
    /// How long, in seconds, the emitter should spawn particles for
    /// </summary>
    EMITTER_SPAWN_TIME,
    
    /// <summary>
    /// How many particles per second the emitter should generate
    /// </summary>
    EMITTER_SPAWN_RATE,
    #endregion

    #region Particle Control field IDs
    /// <summary>
    /// Determines particle movement direction
    /// </summary>
    PARTCILE_DISPERSION_TYPE,
    
    /// <summary>
    /// For direction dispersion, the x component of the direction vector
    /// </summary>
    PARTICLE_DISPERSION_DIR_X,
    
    /// <summary>
    /// For direction dispersion, the y component of the direction vector
    /// </summary>
    PARTICLE_DISPERSION_DIR_Y,
    
    /// <summary>
    /// For direction dispersion, the z component of the direction vector
    /// </summary>
    PARTICLE_DISPERSION_DIR_Z,
    
    /// <summary>
    /// For partial sphere dispersion, the min x angle for the region to use
    /// for particle direction, in degrees in the range [-180,180]
    /// </summary>
    PARTICLE_DISPERSION_PAR_SPHERE_MIN_X,
    
    /// <summary>
    /// For partial sphere dispersion, the max x angle for the region to use
    /// for particle direction, in degrees in the range [-180,180]
    /// </summary>
    PARTICLE_DISPERSION_PAR_SPHERE_MAX_X,
    
    /// <summary>
    /// For partial sphere dispersion, the min y angle for the region to use
    /// for particle direction, in degrees in the range [-180,180]
    /// </summary>
    PARTICLE_DISPERSION_PAR_SPHERE_MIN_Y,
    
    /// <summary>
    /// For partial sphere dispersion, the max y angle for the region to use
    /// for particle direction, in degrees in the range [-180,180]
    /// </summary>
    PARTICLE_DISPERSION_PAR_SPHERE_MAX_Y,
    
    /// <summary>
    /// Shape of the particles
    /// </summary>
    PARTICLE_SHAPE,
    
    /// <summary>
    /// How long in ms individual particles should live for in seconds
    /// </summary>
    PARTICLE_LIFETIME,
    
    /// <summary>
    /// Minimium speed, in units/second, that particles can move at
    /// </summary>
    PARTICLE_MIN_SPEED,
    
    /// <summary>
    /// Maximium speed, in units/second, that particles can move at
    /// </summary>
    PARTICLE_MAX_SPEED,
    
    /// <summary>
    /// Size of the particles in units.  Must be non-negative
    /// </summary>
    PARTICLE_SIZE,
    
    /// <summary>
    /// x component of the particle acceleration vector, in units/second^2
    /// </summary>
    PARTICLE_ACC_X,
    
    /// <summary>
    /// y component of the particle acceleration vector, in units/second^2
    /// </summary>
    PARTICLE_ACC_Y,
    
    /// <summary>
    /// z component of the particle acceleration vector, in units/second^2
    /// </summary>
    PARTICLE_ACC_Z,
    #endregion

    #region Particle Coloring IDs
    /// <summary>
    /// How particles should have their color assigned
    /// </summary>
    COLORING_MODE,
    
    /// <summary>
    /// Path to the texture to use
    /// </summary>
    COLORING_TEXTURE_PATH,
    
    /// <summary>
    /// The entire color set data
    /// </summary>
    COLORING_SET,
    
    /// <summary>
    /// The entire color sequence data, in order
    /// </summary>
    COLORING_SEQ,
    
    /// <summary>
    /// Whether colors of a color sequence should be interpolated between the
    /// current and next value (1), or if the color should be just the current
    /// color (0)
    /// </summary>
    COLORING_SEQ_INTERPOLATE,
    #endregion
    
    /// <summary>
    /// not a parameter id, but a place holder for 1 above the highest valid
    /// parameter id
    /// </summary>
    PARAM_ID_MAX,
  }
}
