﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Xna.Framework;

namespace XNA_Renderer
{
  /// <summary>
  /// Orbiting style camera
  /// </summary>
  class Camera
  {
    /// <summary>
    /// Creates a camera based on the specified data
    /// </summary>
    /// <param name="aspect_ratio">
    /// Aspect ratio for the camera
    /// </param>
    /// <param name="lookat">
    /// Where the camera should be looking in world space
    /// </param>
    /// <param name="range">
    /// distance the camera's position should be from the look at
    /// </param>
    /// <param name="yaw">
    /// angle of the current position of the orbit, in radians in the range
    /// [-pi,pi]
    /// </param>
    /// <param name="pitch">
    /// angle to pitch the camera up/down by, in radians in the range
    /// [-pi/2,pi/2]
    /// </param>
    public Camera(float aspect_ratio,Vector3 lookat,float range,float yaw,
      float pitch)
    {
      m_look_at = lookat;
      m_range   = range;
      m_yaw     = yaw;
      m_pitch   = pitch;

      Projection = Matrix.CreatePerspectiveFieldOfView(MathHelper.PiOver4,
        aspect_ratio,.1f,100);
      
      Recalc();
    }
    
    /// <summary>
    /// Sets the x coordinate in world space of what the camera should be
    /// looking at
    /// </summary>
    /// <param name="x">
    /// x coordinate in world space
    /// </param>
    public void SetLookAtX(float x)
    {
      m_look_at.X = x;
      
      Recalc();
    }
    
    /// <summary>
    /// Sets the y coordinate in world space of what the camera should be
    /// looking at
    /// </summary>
    /// <param name="y">
    /// x coordinate in world space
    /// </param>
    public void SetLookAtY(float y)
    {
      m_look_at.Y = y;
      
      Recalc();
    }
    
    /// <summary>
    /// Sets the z coordinate in world space of what the camera should be
    /// looking at
    /// </summary>
    /// <param name="z">
    /// x coordinate in world space
    /// </param>
    public void SetLookAtZ(float z)
    {
      m_look_at.Z = z;
      
      Recalc();
    }
    
    /// <summary>
    /// Sets the distance the camera's position should be from the look at
    /// </summary>
    /// <param name="range">
    /// distance the camera's position should be from the look at
    /// </param>
    public void SetOrbitRange(float range)
    {
      m_range = range;
      
      Recalc();
    }
    
    /// <summary>
    /// Sets the yaw for determining the camera's position relative to the look
    /// at
    /// </summary>
    /// <param name="yaw">
    /// angle of the current position of the orbit, in radians in the range
    /// [-pi,pi]
    /// </param>
    public void SetOrbitYaw(float yaw)
    {
      m_yaw = yaw;
      
      Recalc();
    }
    
    /// <summary>
    /// Sets the pitch for determining the camera's position relative to the
    /// look at
    /// </summary>
    /// <param name="pitch">
    /// angle to pitch the camera up/down by, in radians in the range
    /// [-pi/2,pi/2]
    /// </param>
    public void SetOrbitPitch(float pitch)
    {
      m_pitch = pitch;
      
      Recalc();
    }

    /// <summary>
    /// View matrix
    /// </summary>
    public Matrix View { get; private set; }

    /// <summary>
    /// Projection matrix
    /// </summary>
    public Matrix Projection { get; private set; }
    
    /// <summary>
    /// Calculated position of the camera
    /// </summary>
    public Vector3 Position { get; private set; }
    
    /// <summary>
    /// Calculated unit vector for direction the camera is pointing
    /// </summary>
    public Vector3 Direction { get; private set; }
    
    /// <summary>
    /// Calculated unit vector for the up direction of the camera
    /// </summary>
    public Vector3 Up { get; private set; }
    
    /// <summary>
    /// Calculated unit vector that points to the right of the direction the
    /// camera is facing
    /// <remarks>
    /// Useful for billboarding
    /// </remarks>
    /// </summary>
    public Vector3 Right { get; private set; }
    
    /// <summary>
    /// Causes the calculated fields to have their values recalculated based on
    /// the current non-calculated fields
    /// </summary>
    private void Recalc()
    {
      Matrix convert =
        Matrix.CreateTranslation(new Vector3(0,0,m_range))
        * Matrix.CreateFromYawPitchRoll(m_yaw,m_pitch,0);

      Position  = Vector3.Transform(m_look_at,convert);
      Direction = Vector3.Normalize(m_look_at - Position);
      Up        = Vector3.TransformNormal(Vector3.UnitY,convert);
      Right     = Vector3.Cross(Up,Direction);
      
      View = Matrix.CreateLookAt(Position,m_look_at,Up);
    }

    /// <summary>
    /// Where the camera should be looking in world space
    /// </summary>
    private Vector3 m_look_at;
    
    /// <summary>
    /// distance the camera's position should be from the look at
    /// </summary>
    private float m_range;
    
    /// <summary>
    /// angle of the current position of the orbit, in radians in the range
    /// [-pi,pi]
    /// </summary>
    private float m_yaw;
    
    /// <summary>
    /// angle to pitch the camera up/down by, in radians in the range
    /// [-pi/2,pi/2]
    /// </summary>
    private float m_pitch;
  }
}
