using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using System.Net;
using System.Net.Sockets;
using System.IO;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Media;

namespace XNA_Renderer
{
  /// <summary>
  /// This is the main type for your game
  /// </summary>
  public class GameMain : Microsoft.Xna.Framework.Game
  {
    GraphicsDeviceManager graphics;
    SpriteBatch spriteBatch;

    public GameMain()
    {
      graphics = new GraphicsDeviceManager(this);
      Content.RootDirectory = "Content";
      m_term_complete = false;
      
      // setup delegates for updating the particle system
      m_updates =
        new Dictionary<NetworkCommon.ParameterIDs,UpdateParamFloat>();
      m_updates.Add(NetworkCommon.ParameterIDs.CAM_LOOKAT_X,
        (x) => m_cam.SetLookAtX(x));
      m_updates.Add(NetworkCommon.ParameterIDs.CAM_LOOKAT_Y,
        (y) => m_cam.SetLookAtY(y));
      m_updates.Add(NetworkCommon.ParameterIDs.CAM_LOOKAT_Z,
        (z) => m_cam.SetLookAtZ(z));
      m_updates.Add(NetworkCommon.ParameterIDs.CAM_ORBIT_RANGE,
        (range) => m_cam.SetOrbitRange(range));
      m_updates.Add(NetworkCommon.ParameterIDs.CAM_ORBIT_YAW,
        (yaw) => m_cam.SetOrbitYaw(MathHelper.ToRadians(yaw)));
      m_updates.Add(NetworkCommon.ParameterIDs.CAM_ORBIT_PITCH,
        (pitch) => m_cam.SetOrbitPitch(MathHelper.ToRadians(pitch)));
      m_updates.Add(NetworkCommon.ParameterIDs.EMITTER_SHAPE,
        (id) => m_particle.Emitter.Shape = 
          m_emitters[(NetworkCommon.EmitterShapes)id]);
      m_updates.Add(NetworkCommon.ParameterIDs.EMITTER_POS_X,
        (x) => m_particle.Emitter.SetPositionX(x));
      m_updates.Add(NetworkCommon.ParameterIDs.EMITTER_POS_Y,
        (y) => m_particle.Emitter.SetPositionY(y));
      m_updates.Add(NetworkCommon.ParameterIDs.EMITTER_POS_Z,
        (z) => m_particle.Emitter.SetPositionZ(z));
      m_updates.Add(NetworkCommon.ParameterIDs.EMITTER_ROT_YAW,
        (yaw) => m_particle.Emitter.Yaw = MathHelper.ToRadians(yaw));
      m_updates.Add(NetworkCommon.ParameterIDs.EMITTER_ROT_PITCH,
        (pitch) => m_particle.Emitter.Pitch = MathHelper.ToRadians(pitch));
      m_updates.Add(NetworkCommon.ParameterIDs.EMITTER_ROT_ROLL,
        (roll) => m_particle.Emitter.Roll = MathHelper.ToRadians(roll));
      m_updates.Add(NetworkCommon.ParameterIDs.EMITTER_SCALE_X,
        (x) => m_particle.Emitter.SetScaleX(x));
      m_updates.Add(NetworkCommon.ParameterIDs.EMITTER_SCALE_Y,
        (y) => m_particle.Emitter.SetScaleY(y));
      m_updates.Add(NetworkCommon.ParameterIDs.EMITTER_SCALE_Z,
        (z) => m_particle.Emitter.SetScaleZ(z));
      m_updates.Add(NetworkCommon.ParameterIDs.EMITTER_SPAWN_TIME,
        (time) => m_particle.Emitter.SpawnTime = time);
      m_updates.Add(NetworkCommon.ParameterIDs.EMITTER_SPAWN_RATE,
        (rate) =>
          {
            m_particle.Emitter.SpawnRate = rate;
            m_particle.UpdateMaxParticles();
          });
      m_updates.Add(NetworkCommon.ParameterIDs.PARTCILE_DISPERSION_TYPE,
        (type) => m_particle.ActiveDispersion =
          (NetworkCommon.ParticleDispersion)type);
      m_updates.Add(NetworkCommon.ParameterIDs.PARTICLE_DISPERSION_DIR_X,
        (x) =>
          ((ParticleSystem.DispersionControl.Directional)
          m_particle.Dispersions[NetworkCommon.ParticleDispersion.DIRECTIONAL])
          .SetDirX(x));
      m_updates.Add(NetworkCommon.ParameterIDs.PARTICLE_DISPERSION_DIR_Y,
        (y) =>
          ((ParticleSystem.DispersionControl.Directional)
          m_particle.Dispersions[NetworkCommon.ParticleDispersion.DIRECTIONAL])
          .SetDirY(y));
      m_updates.Add(NetworkCommon.ParameterIDs.PARTICLE_DISPERSION_DIR_Z,
        (z) =>
          ((ParticleSystem.DispersionControl.Directional)
          m_particle.Dispersions[NetworkCommon.ParticleDispersion.DIRECTIONAL])
          .SetDirZ(z));
      m_updates.Add(
        NetworkCommon.ParameterIDs.PARTICLE_DISPERSION_PAR_SPHERE_MIN_X,
        (x) =>
          ((ParticleSystem.DispersionControl.PartialSphere)
          m_particle.Dispersions
          [NetworkCommon.ParticleDispersion.PARTIAL_SPHERE])
          .MinX = MathHelper.ToRadians(x));
      m_updates.Add(
        NetworkCommon.ParameterIDs.PARTICLE_DISPERSION_PAR_SPHERE_MAX_X,
        (x) =>
          ((ParticleSystem.DispersionControl.PartialSphere)
          m_particle.Dispersions
          [NetworkCommon.ParticleDispersion.PARTIAL_SPHERE])
          .MaxX = MathHelper.ToRadians(x));
      m_updates.Add(
        NetworkCommon.ParameterIDs.PARTICLE_DISPERSION_PAR_SPHERE_MIN_Y,
        (y) =>
          ((ParticleSystem.DispersionControl.PartialSphere)
          m_particle.Dispersions
          [NetworkCommon.ParticleDispersion.PARTIAL_SPHERE])
          .MinY = MathHelper.ToRadians(y));
      m_updates.Add(
        NetworkCommon.ParameterIDs.PARTICLE_DISPERSION_PAR_SPHERE_MAX_Y,
        (y) =>
          ((ParticleSystem.DispersionControl.PartialSphere)
          m_particle.Dispersions
          [NetworkCommon.ParticleDispersion.PARTIAL_SPHERE])
          .MaxY = MathHelper.ToRadians(y));
      m_updates.Add(NetworkCommon.ParameterIDs.PARTICLE_SHAPE,
        (shape) =>
          {
            m_particle.ParticleSettings.Shape =
              (NetworkCommon.ParticleShapes)shape;
            m_particle.UpdateParticleShape();
          });
      m_updates.Add(NetworkCommon.ParameterIDs.PARTICLE_LIFETIME,
        (time) =>
          {
            m_particle.ParticleSettings.Lifetime = time;
            m_particle.UpdateMaxParticles();
          });
      m_updates.Add(NetworkCommon.ParameterIDs.PARTICLE_MIN_SPEED,
        (speed) => m_particle.ParticleSettings.MinSpeed = speed);
      m_updates.Add(NetworkCommon.ParameterIDs.PARTICLE_MAX_SPEED,
        (speed) => m_particle.ParticleSettings.MaxSpeed = speed);
      m_updates.Add(NetworkCommon.ParameterIDs.PARTICLE_SIZE,
        (size) => m_particle.ParticleSettings.Size = size);
      m_updates.Add(NetworkCommon.ParameterIDs.PARTICLE_ACC_X,
        (x) => m_particle.ParticleSettings.SetAccX(x));
      m_updates.Add(NetworkCommon.ParameterIDs.PARTICLE_ACC_Y,
        (y) => m_particle.ParticleSettings.SetAccY(y));
      m_updates.Add(NetworkCommon.ParameterIDs.PARTICLE_ACC_Z,
        (z) => m_particle.ParticleSettings.SetAccZ(z));
      m_updates.Add(NetworkCommon.ParameterIDs.COLORING_MODE,
        (mode) => m_particle.ColoringSettings.ActiveMode =
          (NetworkCommon.ParticleColoringMode)mode);
      m_updates.Add(NetworkCommon.ParameterIDs.COLORING_SEQ_INTERPOLATE,
        (val) => m_particle.ColoringSettings.Interpolate = val > .5f);
    }

    /// <summary>
    /// Allows the game to perform any initialization it needs to before starting to run.
    /// This is where it can query for any required services and load any non-graphic
    /// related content.  Calling base.Initialize will enumerate through any components
    /// and initialize them as well.
    /// </summary>
    protected override void Initialize()
    {
      // create the default camera
      m_cam = new Camera(GraphicsDevice.Viewport.AspectRatio,
        new Vector3(0,0,0),10,0,0);

      // create the various emitters
      m_emitters = new Dictionary<NetworkCommon.EmitterShapes,
        ParticleSystem.EmitterShapes.EmitterShapeBase>();
      m_emitters.Add(NetworkCommon.EmitterShapes.POINT,
        new ParticleSystem.EmitterShapes.Point());
      m_emitters.Add(NetworkCommon.EmitterShapes.LINE,
        new ParticleSystem.EmitterShapes.Line());
      m_emitters.Add(NetworkCommon.EmitterShapes.SQUARE,
        new ParticleSystem.EmitterShapes.Square());
      m_emitters.Add(NetworkCommon.EmitterShapes.SQUARE_PERIMETER,
        new ParticleSystem.EmitterShapes.SquarePerimeter());
      m_emitters.Add(NetworkCommon.EmitterShapes.CIRCLE,
        new ParticleSystem.EmitterShapes.Circle());
      m_emitters.Add(NetworkCommon.EmitterShapes.RING,
        new ParticleSystem.EmitterShapes.Ring());
      m_emitters.Add(NetworkCommon.EmitterShapes.SPHERE,
        new ParticleSystem.EmitterShapes.Sphere());
      m_emitters.Add(NetworkCommon.EmitterShapes.HEMISPHERE,
        new ParticleSystem.EmitterShapes.Hemisphere());
      
      // create the cancelation token
      m_term = new CancellationTokenSource();
      
      // start the server so clients can connect
      new Thread(() => ServerThread(m_term.Token,this)).Start();

      base.Initialize();
    }

    /// <summary>
    /// Takes care of cleaning up the server thread when the program is
    /// terminating
    /// </summary>
    protected override void OnExiting(object sender,EventArgs args)
    {
      base.OnExiting(sender,args);
      
      // stop the server
      m_term.Cancel();
        
      // wait for the server thread to stop
      while (!m_term_complete)
      {
        Thread.Sleep(10);
      }
        
      // inform the client of the terminate, if there is one
      if (m_client != null)
      {
        NetworkCommon.MsgCloseConnection msg =
          new NetworkCommon.MsgCloseConnection();
        msg.Send(m_client.GetStream());
          
        // close the socket
        lock (this)
        {
          ((IDisposable)m_client).Dispose();
          m_client = null;
        }
      }
    }

    /// <summary>
    /// LoadContent will be called once per game and is the place to load
    /// all of your content.
    /// </summary>
    protected override void LoadContent()
    {
      // Create a new SpriteBatch, which can be used to draw textures.
      spriteBatch = new SpriteBatch(GraphicsDevice);
      
      // prep the particle system type
      ParticleSystem.ParticleSystem.LoadContent(GraphicsDevice,Content);
      
      // create an empty particle system
      m_particle = new ParticleSystem.ParticleSystem(GraphicsDevice);
      
      // disable backface culling
      RasterizerState rast = new RasterizerState();
      rast.CullMode = CullMode.None;
      GraphicsDevice.RasterizerState = rast;
    }

    /// <summary>
    /// UnloadContent will be called once per game and is the place to unload
    /// all content.
    /// </summary>
    protected override void UnloadContent()
    {
      // TODO: Unload any non ContentManager content here
    }

    /// <summary>
    /// Allows the game to run logic such as updating the world,
    /// checking for collisions, gathering input, and playing audio.
    /// </summary>
    /// <param name="gameTime">Provides a snapshot of timing values.</param>
    protected override void Update(GameTime gameTime)
    {
      // Allows the game to exit
      if (GamePad.GetState(PlayerIndex.One).Buttons.Back
        == ButtonState.Pressed
        || Keyboard.GetState().IsKeyDown(Keys.Escape))
      {
        Exit();
      }
      
      // check for any messages from the client
      if (m_client != null)
      {
        while (m_client != null && m_client.Available > 0)
        {
          // get the message
          NetworkCommon.Message msg =
            NetworkCommon.Message.Receive(m_client.GetStream());
        
          // process the message
          switch (msg.GetMessageID())
          {
            case NetworkCommon.MessageIDs.CLOSE_CONNECTION:
              lock (this)
              {
                // client has closed the connection, so clean it up on this
                // side
                ((IDisposable)m_client).Dispose();
                m_client = null;
              }
              break;
              
            case NetworkCommon.MessageIDs.CONNECTION_REJECTED:
              // ignore, clients aren't supposed to send this message
              break;
              
            case NetworkCommon.MessageIDs.SET_SINGLE:
              NetworkCommon.MsgSetSingle single =
                (NetworkCommon.MsgSetSingle)msg;
              SetParam(single.ID,single.Value);
              break;
              
            case NetworkCommon.MessageIDs.SET_STRING:
              NetworkCommon.MsgSetString str =
                (NetworkCommon.MsgSetString)msg;
              SetParam(str.ID,str.Value);
              break;
              
            case NetworkCommon.MessageIDs.COLOR_SET:
              NetworkCommon.MsgColorSet color_set =
                (NetworkCommon.MsgColorSet)msg;
              SetParam(color_set.Value);
              break;
              
            case NetworkCommon.MessageIDs.COLOR_SEQ:
              NetworkCommon.MsgColorSeq color_seq =
                (NetworkCommon.MsgColorSeq)msg;
              SetParam(color_seq.Value);
              break;

            default:
              // ignore
              break;
          }
        }
      }
      
      // skip updates with 0 time
      if (gameTime.ElapsedGameTime.TotalMilliseconds == 0)
      {
        base.Update(gameTime);
        return;
      }

      // update the particle system
      m_particle.Update((float)gameTime.ElapsedGameTime.TotalSeconds);

      base.Update(gameTime);
    }

    /// <summary>
    /// This is called when the game should draw itself.
    /// </summary>
    /// <param name="gameTime">Provides a snapshot of timing values.</param>
    protected override void Draw(GameTime gameTime)
    {
      GraphicsDevice.Clear(Color.CornflowerBlue);

      // draw the particle system
      m_particle.Draw(m_cam);

      base.Draw(gameTime);
    }
    
    /// <summary>
    /// Thread functiion for receiver connections from the input program
    /// </summary>
    /// <param name="token">
    /// Token to use for determining when to end the thread
    /// </param>
    /// <param name="game">
    /// GameMain instance that is running this thread
    /// </param>
    static private void ServerThread(CancellationToken token,GameMain game)
    {
      TcpListener listener = new TcpListener(IPAddress.Any,
        NetworkCommon.Constants.DefaultPort);
      listener.Start();
      
      while (token.IsCancellationRequested == false)
      {
        if (listener.Pending())
        {
          lock (game)
          {
            // if there is no current client, allow this one
            if (game.m_client == null)
            {
              game.m_client = listener.AcceptTcpClient();
            }
            // there is already a client, send an error message back
            else
            {
              using (TcpClient client = listener.AcceptTcpClient())
              using (NetworkStream stream = client.GetStream())
              {
                NetworkCommon.MsgConnectionRejected msg = new
                 NetworkCommon.MsgConnectionRejected();
                msg.Send(stream);
              }
            }
          }
        }
        else
        {
          Thread.Sleep(10);
        }
      }
      
      game.m_term_complete = true;
    }
    
    /// <summary>
    /// Takes care of setting a parameter.
    /// <remarks>
    /// This function exists to try to avoid cluttering the Update function
    /// </remarks>
    /// </summary>
    /// <param name="id">
    /// ID of the parameter to update
    /// </param>
    /// <param name="val">
    /// value to update the parameter to
    /// </param>
    private void SetParam(NetworkCommon.ParameterIDs id,float val)
    {
      UpdateParamFloat update = null;
      if (m_updates.TryGetValue(id,out update))
      {
        update(val);
      }
    }
    
    /// <summary>
    /// Takes care of setting a parameter.
    /// <remarks>
    /// This function exists to try to avoid cluttering the Update function
    /// </remarks>
    /// </summary>
    /// <param name="id">
    /// ID of the parameter to update
    /// </param>
    /// <param name="val">
    /// value to update the parameter to
    /// </param>
    private void SetParam(NetworkCommon.ParameterIDs id,string val)
    {
      if (id == NetworkCommon.ParameterIDs.COLORING_TEXTURE_PATH
        && val != "")
      {
        // send the data to the particle system
        FileStream stream = File.OpenRead(val);
        m_particle.ColoringSettings.Texture =
          Texture2D.FromStream(GraphicsDevice,stream);
      }
      // ignore other ids
    }
    
    /// <summary>
    /// Takes care of setting a parameter.
    /// <remarks>
    /// This function exists to try to avoid cluttering the Update function
    /// </remarks>
    /// </summary>
    /// <param name="val">
    /// value to update the parameter to
    /// </param>
    private void SetParam(NetworkCommon.ColorSetEntry[] val)
    {
      // make sure the entries have a total probability of 100%
      float total = 0;
      foreach (NetworkCommon.ColorSetEntry entry in val)
      {
        total += entry.Chance;
      }
      foreach (NetworkCommon.ColorSetEntry entry in val)
      {
        entry.Chance /= total;
      }
      
      // send the data to the particle system
      m_particle.ColoringSettings.ColorSet = val;
    }
    
    /// <summary>
    /// Takes care of setting a parameter.
    /// <remarks>
    /// This function exists to try to avoid cluttering the Update function
    /// </remarks>
    /// </summary>
    /// <param name="val">
    /// value to update the parameter to
    /// </param>
    private void SetParam(NetworkCommon.ColorSeqEntry[] val)
    {
      // make sure the entries have a total probability of 100%
      float total = 0;
      foreach (NetworkCommon.ColorSeqEntry entry in val)
      {
        total += entry.Life;
      }
      foreach (NetworkCommon.ColorSeqEntry entry in val)
      {
        entry.Life /= total;
      }
      
      // send the data to the particle system
      m_particle.ColoringSettings.ColorSeq = val;
    }
    
    /// <summary>
    /// Delegate type for updating a float param
    /// </summary>
    /// <param name="val">
    /// Value to set the param to
    /// </param>
    private delegate void UpdateParamFloat(float val);
    
    /// <summary>
    /// Map of parameter ids to functions that take care of updating them
    /// </summary>
    /// <remarks>
    /// This exists only for float values since for strings, color sets, and
    /// color sequences there is only 1 function to call for each of those
    /// types and a lookup is not needed
    /// </remarks>
    private Dictionary<NetworkCommon.ParameterIDs,UpdateParamFloat> m_updates;
    
    /// <summary>
    /// particle system to edit
    /// </summary>
    private ParticleSystem.ParticleSystem m_particle;
    
    /// <summary>
    /// map of emitter shape ids to emitter instances.  Due to the way emitters
    /// work, there is no need to have more than one instance of a particular
    /// shape of emitter
    /// </summary>
    private Dictionary<NetworkCommon.EmitterShapes,
      ParticleSystem.EmitterShapes.EmitterShapeBase> m_emitters;
    
    /// <summary>
    /// camera used to render the particle system
    /// </summary>
    private Camera m_cam;
    
    /// <summary>
    /// token used for informing the server thread when to stop
    /// </summary>
    private CancellationTokenSource m_term;
    
    /// <summary>
    /// informs the instance of when its connection thread has finished
    /// </summary>
    private bool m_term_complete;
    
    /// <summary>
    /// connection to the client
    /// </summary>
    /// <remarks>
    /// Setting this should only be done while holding a lock on the
    /// GameMain instance.  Due to how GameMain instances work, setting rather
    /// than all accesses is the only time a lock needs to be held for this.
    /// The helper thread takes care of setting it when not null, and the main
    /// thread disposes of it and sets it to null.  All other access is by the
    /// main thread and is using the instance if not null.
    /// </remarks>
    private TcpClient m_client;
  }
}
