﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

namespace XNA_Renderer.ParticleSystem
{
  /// <summary>
  /// Store data about the various coloring modes
  /// </summary>
  class ColoringSettings
  {
    /// <summary>
    /// Which coloring mode is currently in use
    /// </summary>
    public NetworkCommon.ParticleColoringMode ActiveMode { get; set; }
    
    /// <summary>
    /// For the texture coloring mode, the texture to apply to the particles
    /// </summary>
    public Texture2D Texture { get; set; }
    
    /// <summary>
    /// For the color set coloring mode, the list of colors and their
    /// associated probabilities
    /// </summary>
    public NetworkCommon.ColorSetEntry[] ColorSet { get; set; }
    
    /// <summary>
    /// For color sequence coloring mode, the list of colors in order and the
    /// associated percentage of particle life that entry should be used for
    /// </summary>
    public NetworkCommon.ColorSeqEntry[] ColorSeq { get; set; }
    
    /// <summary>
    /// For the color sequence mode, true if colors should be interpolated
    /// between the current and next color, false if it should just be the
    /// current color
    /// </summary>
    public bool Interpolate { get; set; }
    
    /// <summary>
    /// Determines the color of a particle at its creation time
    /// </summary>
    /// <returns>
    /// the color for the particle
    /// </returns>
    public Vector3 GetInitialColor()
    {
      Vector3 back = Vector3.One;
      
      switch (ActiveMode)
      {
        case NetworkCommon.ParticleColoringMode.COLOR_SET:
          if (ColorSet != null && ColorSet.Length > 0)
          {
            float val   = StaticRandom.Rng.NextFloat();
            float total = 0;
            for (int j = 0; j < ColorSet.Length; ++j)
            {
              if (val <= ColorSet[j].Chance + total)
              {
                back = ColorSet[j].Color;
                break;
              }
                
              total += ColorSet[j].Chance;
            }
          }
          break;
              
        case NetworkCommon.ParticleColoringMode.COLOR_SEQ:
          if (ColorSeq != null && ColorSeq.Length > 0)
          {
            back = ColorSeq[0].Color;
          }
          break;
      }
      
      return back;
    }
    
    /// <summary>
    /// Determines the color of a particle at the specified percentage of its
    /// life
    /// <remarks>
    /// Should only be called when color sequence is the active coloring mode
    /// </remarks>
    /// </summary>
    /// <param name="life">
    /// how far the particle is in its life, in the range (0,1]
    /// </param>
    /// <returns>
    /// color for the particle
    /// </returns>
    public Vector3 GetColor(float life)
    {
      int i = 0;
      float total = 0;
      for (; i < ColorSeq.Length; ++i)
      {
        total += ColorSeq[i].Life;
        
        if (total > life)
        {
          break;
        }
      }
      
      if (Interpolate && ColorSeq.Length > 0 && i != ColorSeq.Length - 1)
      {
        // determine how far between the current and next color sequence
        // entries the current life is, in the range [0,1]
        float mul = (ColorSeq[i].Life + life - total) / ColorSeq[i].Life;
        
        // use the multiplier to determine the color value
        return (1 - mul) * ColorSeq[i].Color + mul * ColorSeq[i + 1].Color;
      }
      else if (ColorSeq.Length > 0)
      {
        return ColorSeq[i].Color;
      }
      else
      {
        return Vector3.One;
      }
    }
  }
}
