﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Xna.Framework;

namespace XNA_Renderer.ParticleSystem
{
  /// <summary>
  /// Stores settings for an emitter
  /// </summary>
  class EmitterSettings
  {
    /// <summary>
    /// Retrieves the position of the emitter in world space
    /// </summary>
    /// <returns>
    /// position of the emitter, in world space
    /// </returns>
    public Vector3 GetPosition()
    {
      return m_pos;
    }
    
    /// <summary>
    /// Retrieves the scale factors for the emitter, in emitter local space
    /// </summary>
    /// <returns>
    /// scale factors for the emitter, in emitter local space
    /// </returns>
    public Vector3 GetScale()
    {
      return m_emitter_scale;
    }
    
    /// <summary>
    /// Sets the x coordinate in world space of the emitter's position
    /// </summary>
    /// <param name="x">
    /// x coordinate in world space
    /// </param>
    public void SetPositionX(float x)
    {
      m_pos.X = x;
    }
    
    /// <summary>
    /// Sets the y coordinate in world space of the emitter's position
    /// </summary>
    /// <param name="y">
    /// x coordinate in world space
    /// </param>
    public void SetPositionY(float y)
    {
      m_pos.Y = y;
    }
    
    /// <summary>
    /// Sets the z coordinate in world space of the emitter's position
    /// </summary>
    /// <param name="z">
    /// x coordinate in world space
    /// </param>
    public void SetPositionZ(float z)
    {
      m_pos.Z = z;
    }
    
    /// <summary>
    /// Sets the x scale component in the emitter's local space
    /// </summary>
    /// <param name="x">
    /// y scale component in emitter local space
    /// </param>
    public void SetScaleX(float x)
    {
      m_emitter_scale.X = x;
    }
    
    /// <summary>
    /// Sets the y scale component in the emitter's local space
    /// </summary>
    /// <param name="y">
    /// y scale component in emitter local space
    /// </param>
    public void SetScaleY(float y)
    {
      m_emitter_scale.Y = y;
    }
    
    /// <summary>
    /// Sets the z scale component in the emitter's local space
    /// </summary>
    /// <param name="z">
    /// z scale component in emitter local space
    /// </param>
    public void SetScaleZ(float z)
    {
      m_emitter_scale.Z = z;
    }
    
    /// <summary>
    /// Emitter that determines the initial position of the particles
    /// </summary>
    public EmitterShapes.EmitterShapeBase Shape { get; set; }
    
    /// <summary>
    /// Yaw for the rotation of the emitter's direction, in radians
    /// </summary>
    public float Yaw { get; set; }
    
    /// <summary>
    /// Pitch for the rotation of the emitter's direction, in radians
    /// </summary>
    public float Pitch { get; set; }
    
    /// <summary>
    /// Roll for the rotation of the emitter's direction, in radians
    /// </summary>
    public float Roll { get; set; }
    
    /// <summary>
    /// Amount of time, in seconds, that the emitter should spawn particles for
    /// </summary>
    public float SpawnTime { get; set; }
    
    /// <summary>
    /// How many particles per second the emitter should generate
    /// </summary>
    public float SpawnRate { get; set; }
    
    /// <summary>
    /// Position of the emitter in world space
    /// </summary>
    private Vector3 m_pos;
    
    /// <summary>
    /// Amount to scale the emitter by, in the emitter's local space
    /// </summary>
    /// <remarks>
    /// Emitter scaling is done by the particle system instead of the emitter
    /// base class so that changing an emitter keeps the same scaling.  The
    /// scaling could be made a static member of the emitter base class, but
    /// if multiple emitters are used, that would force them all to use the
    /// same scaling.
    /// </remarks>
    private Vector3 m_emitter_scale;
  }
}
