﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Content;

namespace XNA_Renderer.ParticleSystem
{
  /// <summary>
  /// Instances take care of managing individual particle systems
  /// </summary>
  class ParticleSystem
  {
    /// <summary>
    /// Allows the particle system type to load any content that it needs to
    /// </summary>
    /// <param name="graphics">
    /// graphics device that will be used by particle system instances
    /// </param>
    /// <param name="content">
    /// content manager to load data from
    /// </param>
    static public void LoadContent(GraphicsDevice graphics,
      ContentManager content)
    {
      // load the effect for rendering
      s_billboard_effect = content.Load<Effect>(
        "ParticleSystem/BillboardInstanceRenderer");
      
      // create the triangle particle vertex and index buffer
      VertexPositionTexture[] verts   = new VertexPositionTexture[3];
      int[]                   indices = new int[3];
      verts[0] = new VertexPositionTexture(new Vector3(0,.5f,0),
        new Vector2(.5f,0));
      verts[1] = new VertexPositionTexture(new Vector3(.5f,-.5f,0),
        Vector2.UnitY);
      verts[2] = new VertexPositionTexture(new Vector3(-.5f,-.5f,0),
        Vector2.One);
      s_vert_tri = new VertexBuffer(graphics,
        VertexPositionTexture.VertexDeclaration,verts.Length,
        BufferUsage.WriteOnly);
      s_vert_tri.SetData(verts);
      indices[0] = 0;
      indices[1] = 1;
      indices[2] = 2;
      s_index_tri = new IndexBuffer(graphics,typeof(int),indices.Length,
        BufferUsage.WriteOnly);
      s_index_tri.SetData(indices);
      
      // create the square particle vertex and index buffer
      verts = new VertexPositionTexture[4];
      verts[0] = new VertexPositionTexture(new Vector3(-.5f,-.5f,0),
        Vector2.One);
      verts[1] = new VertexPositionTexture(new Vector3(-.5f, .5f,0),
        Vector2.UnitX);
      verts[2] = new VertexPositionTexture(new Vector3( .5f, .5f,0),
        Vector2.Zero);
      verts[3] = new VertexPositionTexture(new Vector3( .5f,-.5f,0),
        Vector2.UnitY);
      s_vert_square = new VertexBuffer(graphics,
        VertexPositionTexture.VertexDeclaration,verts.Length,
        BufferUsage.WriteOnly);
      s_vert_square.SetData(verts);
      indices = new int[4];
      indices[0] = 0;
      indices[1] = 1;
      indices[2] = 3;
      indices[3] = 2;
      s_index_square = new IndexBuffer(graphics,typeof(int),indices.Length,
        BufferUsage.WriteOnly);
      s_index_square.SetData(indices);
      
      // setup the particle instance vertex declaration
      VertexElement[] elems = new VertexElement[2];
      elems[0] = new VertexElement(0,VertexElementFormat.Vector4,
        VertexElementUsage.TextureCoordinate,1);
      elems[1] = new VertexElement(sizeof(float) * 4,
        VertexElementFormat.Vector3,VertexElementUsage.TextureCoordinate,2);
      s_instance_dec = new VertexDeclaration(elems);
    }
    
    /// <summary>
    /// Creates particle sysetm.  This should only be called after LoadContent
    /// has been done for the type.
    /// </summary>
    /// <param name="device">
    /// graphics device that this particle system will render to.  Must be the
    /// same as the one given to LoadContent
    /// </param>
    public ParticleSystem(GraphicsDevice device)
    {
      // setup the storage for the settings
      Emitter          = new EmitterSettings();
      ParticleSettings = new ParticleSettings();
      ColoringSettings = new ColoringSettings();
      Dispersions      = new Dictionary<NetworkCommon.ParticleDispersion,
        DispersionControl.DispersionBase>();
      
      // create instances for the various dispersion types
      Dispersions.Add(NetworkCommon.ParticleDispersion.DIRECTIONAL,
        new DispersionControl.Directional());
      Dispersions.Add(NetworkCommon.ParticleDispersion.SPHERICAL,
        new DispersionControl.Spherical());
      Dispersions.Add(NetworkCommon.ParticleDispersion.PARTIAL_SPHERE,
        new DispersionControl.PartialSphere());
      
      // store the graphics device
      m_graphics = device;
      
      // create the array for the vertex bindings for the instance renderer
      m_binding = new VertexBufferBinding[2];
      
      // no additional spawn time yet
      m_additonal_spawn_time = 0;
      
      // emitter has not produced particles yet
      m_runtime = 0;
    }
    
    /// <summary>
    /// Informs the particle system that the spawn rate or particle lifetime
    /// (the variables that affect the max number of particles existing at any
    /// given moment) have been updated
    /// </summary>
    /// <remarks>
    /// This is just to avoid having to calculate the max number of particles
    /// and comparing to the ring buffer size every update
    /// </remarks>
    public void UpdateMaxParticles()
    {
      m_additonal_spawn_time = 0;
      m_runtime = 0;
      
      if (Emitter.SpawnRate > 0 && ParticleSettings.Lifetime > 0)
      {
        // calculate the max number of particles that can exist at any given
        // time
        int max =
          (int)Math.Round(Emitter.SpawnRate * ParticleSettings.Lifetime);
      
        // (re)create the ring buffer based on the size
        if (m_particles == null)
        {
          m_particles = new RingBuffer<ParticleData>(max);
          m_instance_data = new ParticleInstance[max];
          m_instance = new VertexBuffer(m_graphics,s_instance_dec,max,
            BufferUsage.WriteOnly);
          m_binding[1] = new VertexBufferBinding(m_instance,0,1);
        }
        else if (m_particles.Data.Length != max)
        {
          m_particles = new RingBuffer<ParticleData>(max,m_particles);
          m_instance_data = new ParticleInstance[max];
          
          if (m_instance != null)
          {
            m_instance.Dispose();
          }
          m_instance = new VertexBuffer(m_graphics,s_instance_dec,max,
            BufferUsage.WriteOnly);
          m_binding[1] = new VertexBufferBinding(m_instance,0,1);
        }
      }
      else
      {
        m_particles = null;
        if (m_instance != null)
        {
          m_instance.Dispose();
          m_instance = null;
        }
      }
    }
    
    /// <summary>
    /// Informs the particle system that particle shape has been changed
    /// </summary>
    public void UpdateParticleShape()
    {
      if (ParticleSettings.Shape != NetworkCommon.ParticleShapes.TRIANGLE)
      {
        m_binding[0] = new VertexBufferBinding(s_vert_square,0);
      }
      else
      {
        m_binding[0] = new VertexBufferBinding(s_vert_tri,0);
      }
    }
    
    /// <summary>
    /// Updates the particle system based on time
    /// </summary>
    /// <param name="seconds">
    /// Number of seconds to update the particle system by
    /// <remarks>
    /// It is pointless to call this function when this value is 0 aside from
    /// wasting performance
    /// </remarks>
    /// </param>
    public void Update(float seconds)
    {
      if (m_particles != null)
      {
        // update existing particles
        UpdateParticles(seconds);
        
        // create new particles, if the emitter is still active
        if (m_runtime <= Emitter.SpawnTime)
        {
          CreateParticles(seconds);
        }
        
        // update the instance buffer
        UpdateInstanceBuffer();
        
        // update how long the emitter has been running for
        m_runtime += seconds;
        
        // if the emitter has been idle long enough, restart it
        if (m_runtime > Emitter.SpawnTime + ParticleSettings.Lifetime + 1)
        {
          m_runtime = 0;
        }
      }
    }
    
    /// <summary>
    /// Draws the particle system
    /// </summary>
    /// <param name="cam">
    /// Camera to use for rendering the particle system
    /// </param>
    public void Draw(Camera cam)
    {
      if (m_particles != null && m_particles.InUse > 0)
      {
        // save the old blend and depth stencil states
        BlendState        blend_state = m_graphics.BlendState;
        DepthStencilState depth_state = m_graphics.DepthStencilState;
      
        // set the blend and depth stencil states for alpha blending
        m_graphics.BlendState        = BlendState.AlphaBlend;
        m_graphics.DepthStencilState = DepthStencilState.DepthRead;
        
        Vector4 cam_up    = new Vector4(cam.Up,0);
        Vector4 cam_right = new Vector4(cam.Right,0);
        
        s_billboard_effect.Parameters["vp"].SetValue(
          cam.View * cam.Projection);
        s_billboard_effect.Parameters["cam_up"].SetValue(cam_up);
        s_billboard_effect.Parameters["cam_right"].SetValue(cam_right);
        s_billboard_effect.Parameters["scale"].SetValue(ParticleSettings.Size);
        s_billboard_effect.Parameters["particle_texture"].SetValue(
          ColoringSettings.Texture);
        s_billboard_effect.Parameters["use_texture"].SetValue(
          ColoringSettings.ActiveMode
          == NetworkCommon.ParticleColoringMode.TEXTURE ? 1 : 0);
        m_graphics.Indices =
          ParticleSettings.Shape != NetworkCommon.ParticleShapes.TRIANGLE
          ? s_index_square : s_index_tri;
        foreach (EffectPass pass in s_billboard_effect.CurrentTechnique.Passes)
        {
          pass.Apply();
          
          // do the instance rendering
          int verts = m_binding[0].VertexBuffer.VertexCount;
          m_graphics.SetVertexBuffers(m_binding);
          m_graphics.DrawInstancedPrimitives(PrimitiveType.TriangleStrip,0,
            0,verts,0,verts - 2,m_particles.InUse);
          
          // clear the vertex buffers so that Update can change the data in the
          // instance vertex buffer
          m_graphics.SetVertexBuffers(null);
        }
        
        // restore the old blend and depth stencil states
        m_graphics.BlendState        = blend_state;
        m_graphics.DepthStencilState = depth_state;
      }
    }
    
    /// <summary>
    /// Settings for the emitter
    /// </summary>
    public EmitterSettings Emitter { get; private set; }
    
    /// <summary>
    /// Map of particle dispersion type to dispersion control instance
    /// </summary>
    /// <remarks>
    /// a map is used instead of a single instance so that when the editor
    /// switches dispersion types it doesn't need to re-send all the data
    /// associated with that type
    /// </remarks>
    public Dictionary<NetworkCommon.ParticleDispersion,
      DispersionControl.DispersionBase> Dispersions { get; private set; }
    
    /// <summary>
    /// ID of which particle dispersion control is the currently active one
    /// </summary>
    public NetworkCommon.ParticleDispersion ActiveDispersion { get; set; }
    
    /// <summary>
    /// Settings for the particles
    /// </summary>
    public ParticleSettings ParticleSettings { get; private set; }
    
    /// <summary>
    /// Coloring settings for the particles
    /// </summary>
    public ColoringSettings ColoringSettings { get; private set; }
    
    /// <summary>
    /// Creates new particles
    /// </summary>
    /// <param name="seconds">
    /// amount of time, in seconds, to use for calculating the number of
    /// particles to create
    /// </param>
    private void CreateParticles(float seconds)
    {
      // determine how many particles to create
      int num = (int)(Emitter.SpawnRate * (seconds + m_additonal_spawn_time));
      
      if (num > 0)
      {
        m_additonal_spawn_time = 0;
        
        // create a matrix to convert emitter local space to world space
        Matrix rot = 
          Matrix.CreateFromYawPitchRoll(Emitter.Yaw,Emitter.Pitch,
          Emitter.Roll);
        Matrix convert =
          Matrix.CreateScale(Emitter.GetScale())
          * rot * Matrix.CreateTranslation(Emitter.GetPosition());
        
        // create the particles
        for (int i = 0; i < num; ++i)
        {
          ParticleData part;
          float speed = StaticRandom.Rng.NextFloat()
            * (ParticleSettings.MaxSpeed - ParticleSettings.MinSpeed)
            + ParticleSettings.MinSpeed;
          part.Color = ColoringSettings.GetInitialColor();
          part.Position = Vector4.Transform(Emitter.Shape.SpawnPosition(),
            convert);
          part.Velocity = Vector4.Transform(
            Dispersions[ActiveDispersion].GetDirection() * speed,rot);
          part.ExistedFor = 0;
          
          m_particles.Add(part);
        }
      }
      else
      {
        m_additonal_spawn_time += seconds;
      }
    }
    
    /// <summary>
    /// Updates existing particles
    /// </summary>
    /// <param name="seconds">
    /// number of seconds to update the particles by
    /// </param>
    private void UpdateParticles(float seconds)
    {
      // update the particles and put their data into the instance buffer
      Vector4 acceleration = ParticleSettings.GetAcc() * seconds;
      int rem_cnt = 0;
      for (int i = 0; i < m_particles.InUse; ++i)
      {
        int index = m_particles.ToArrayIndex(i);
        
        // update how long the particle has existed for
        m_particles.Data[index].ExistedFor += seconds;
        
        // if the particle is still alive, perform the update
        if (m_particles.Data[index].ExistedFor < ParticleSettings.Lifetime)
        {
          // update the particle velocity based on acceleration
          m_particles.Data[index].Velocity += acceleration;
          
          // update the particle position
          m_particles.Data[index].Position +=
            m_particles.Data[index].Velocity * seconds;
          
          // if needed, update the color as well
          if (ColoringSettings.ActiveMode
            == NetworkCommon.ParticleColoringMode.COLOR_SEQ)
          {
            float life = m_particles.Data[index].ExistedFor
              / ParticleSettings.Lifetime;
            m_particles.Data[index].Color = ColoringSettings.GetColor(life);
          }
        }
        // otherwise, discard the particle
        else
        {
          ++rem_cnt;
        }
      }
      m_particles.Free(rem_cnt);
    }
    
    /// <summary>
    /// Updates the instance buffer based on the current particle data
    /// </summary>
    private void UpdateInstanceBuffer()
    {
      for (int i = 0; i < m_particles.InUse; ++i)
      {
        // put the data in the instance buffer
        int index = m_particles.ToArrayIndex(i);
        m_instance_data[i].Color = m_particles.Data[index].Color;
        m_instance_data[i].Position = m_particles.Data[index].Position;
      }
      m_instance.SetData(m_instance_data);
    }
    
    /// <summary>
    /// Struct for containing the instance rendering data for a particle
    /// </summary>
    private struct ParticleInstance
    {
      /// <summary>
      /// Position of the particle in world space
      /// </summary>
      public Vector4 Position;
      
      /// <summary>
      /// RGB color for the particle
      /// </summary>
      public Vector3 Color;
    };
    
    /// <summary>
    /// Struct for containing the non-rendering data for a particle
    /// </summary>
    private struct ParticleData
    {
      /// <summary>
      /// position of the particle in world space
      /// </summary>
      public Vector4 Position;
      
      /// <summary>
      /// velocity vector in world space
      /// </summary>
      public Vector4 Velocity;
      
      /// <summary>
      /// how long, in seconds, the particle has existed for
      /// </summary>
      public float ExistedFor;
      
      /// <summary>
      /// color of the particle
      /// </summary>
      public Vector3 Color;
    };

    /// <summary>
    /// vertex buffer for the instance data
    /// </summary>
    private VertexBuffer m_instance;
    
    /// <summary>
    /// vertex buffer bindings for the particle geomerty and instance data
    /// </summary>
    private VertexBufferBinding[] m_binding;
    
    /// <summary>
    /// graphics device to render to
    /// </summary>
    private GraphicsDevice m_graphics;
    
    /// <summary>
    /// data about the particles that the particle instance renderer doesn't
    /// need
    /// </summary>
    private RingBuffer<ParticleData> m_particles;
    
    /// <summary>
    /// graphical data for each particle that is alive
    /// </summary>
    /// <remarks>
    /// this being a member variable is to avoid repeated allocation during
    /// Update.  Instead this is allocated to the max number of particles that
    /// can be alive at any given time and only the first m_alive contain the
    /// needed data.
    /// </remarks>
    private ParticleInstance[] m_instance_data;
    
    /// <summary>
    /// Time, in seconds, to add to the amount of time for spawning particles
    /// <remarks>
    /// This allows for spawning particles at a rate less than the rate Update
    /// is called
    /// </remarks>
    /// </summary>
    private float m_additonal_spawn_time;
    
    /// <summary>
    /// keeps track of how long, in seconds, the emitter has been producing
    /// particles for
    /// </summary>
    private float m_runtime;

    #region Static_vars
    /// <summary>
    /// vertex buffer for triangle particles
    /// </summary>
    static private VertexBuffer s_vert_tri;
    
    /// <summary>
    /// vertex buffer for square particles
    /// </summary>
    static private VertexBuffer s_vert_square;
    
    /// <summary>
    /// index buffer for triangle particles
    /// </summary>
    static private IndexBuffer s_index_tri;
    
    /// <summary>
    /// index buffer for square particles
    /// </summary>
    static private IndexBuffer s_index_square;
    
    /// <summary>
    /// vertex declaration for the particle instance data
    /// </summary>
    static private VertexDeclaration s_instance_dec;
    
    /// <summary>
    /// Instanced renderer for the particles
    /// </summary>
    static private Effect s_billboard_effect;
    #endregion // Static_vars
  }
}
