﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace XNA_Renderer.ParticleSystem
{
  /// <summary>
  /// Fixed length ring buffer collection that is specialized for performance
  /// with the ParticleSystem class
  /// </summary>
  /// <typeparam name="T">
  /// Type of elements in the ring buffer
  /// </typeparam>
  class RingBuffer<T> : IEnumerable<T>, IEnumerable
  {
    /// <summary>
    /// Creates an empty ring buffer with the specified number of elements
    /// </summary>
    /// <param name="length">
    /// number of elements in the ring buffer
    /// </param>
    public RingBuffer(int length)
    {
      // create the storage
      Data = new T[length];
      
      // init the start index and number of elements in use
      InUse = 0;
      Start = 0;
      End   = 0;
    }
    
    /// <summary>
    /// Creates a ring buffer that initially has active elements of cpy.
    /// <remarks>
    /// If length is less than the length of cpy, then only the first length
    /// elements are copied
    /// </remarks>
    /// </summary>
    /// <param name="length">
    /// length of the new ring buffer
    /// </param>
    /// <param name="cpy">
    /// ring buffer to copy data from
    /// </param>
    public RingBuffer(int length,RingBuffer<T> cpy)
    :this(length)
    {
      int cnt = (length < cpy.InUse) ? length : cpy.InUse;
      for (int i = 0; i < cnt; i++)
      {
        Add(cpy.Data[cpy.ToArrayIndex(i)]);
      }
    }
    
    /// <summary>
    /// Converts a ring buffer index to an index into the array Data
    /// </summary>
    /// <remarks>
    /// This is needed due to the particle system using structs as the type of
    /// the elements otherwise an enumerator could be used to update the values
    /// of the elements
    /// </remarks>
    /// <param name="i">
    /// ring buffer index
    /// </param>
    /// <returns>
    /// index into the Data array
    /// </returns>
    public int ToArrayIndex(int i)
    {
      return (Start + i) % Data.Length;
    }
    
    /// <summary>
    /// Adds the specified element to the ring buffer
    /// <remarks>
    /// This does not check for the buffer already being full since this
    /// particular implementation is intended for the particle system to keep
    /// track of a number of particles that was given to the constructor
    /// </remarks>
    /// </summary>
    /// <param name="val">
    /// value to add
    /// </param>
    public void Add(T val)
    {
      // store the data
      Data[End] = val;
      
      // update the end index
      End = (End + 1) % Data.Length;

      // update the number of elements in use
      ++InUse;
    }
    
    /// <summary>
    /// Removes the first cnt elements starting at Start's index
    /// <remarks>
    /// This does not check that there are at least cnt elements in use
    /// </remarks>
    /// </summary>
    /// <param name="cnt">
    /// Number of elements to remove from the ring buffer
    /// </param>
    public void Free(int cnt)
    {
      Start = (Start + cnt) % Data.Length;
      InUse -= cnt;
      
      // note: the particle system is using this class on structs, so there is
      // no need to null out the entry in the ring buffer
    }
    
    /// <summary>
    /// Retrieves an enumerator that will go over all the current elements
    /// </summary>
    /// <returns>
    /// enumerator for the current elements
    /// </returns>
    public IEnumerator<T> GetEnumerator()
    {
      int limit = InUse;
      for (int i = 0; i < limit; ++i)
      {
        yield return Data[(i + Start) % Data.Length];
      }
    }
    
    /// <summary>
    /// Non-generic version of GetEnumerator above
    /// </summary>
    IEnumerator IEnumerable.GetEnumerator()
    {
      return GetEnumerator();
    }
    
    /// <summary>
    /// storage of elements in the ring buffer
    /// </summary>
    public T[] Data { get; private set; }
    
    /// <summary>
    /// number of elements in Data that are in use
    /// </summary>
    public int InUse { get; private set; }
    
    /// <summary>
    /// index of the first element in Data that is in use
    /// <remarks>
    /// Should only be used if InUse is > 0
    /// </remarks>
    /// </summary>
    public int Start { get; private set; }
    
    /// <summary>
    /// Index of the last elementin Data that is in use
    /// <remarks>
    /// Should only be used if InUse is > 0
    /// </remarks>
    /// </summary>
    public int End { get; private set; }
  }
}
